"""
========================================
  Introduction to Computer Programming
========================================

-----------------------------------------------
  Programming Assignment 8
  Conway's Game of Life

  <YOUR NAME(s)>

  ... status ...
-----------------------------------------------
"""


# ...

from grid_graphics import (open_window, close_window, get_dimensions,
                           clear, paint, update,
                           set_click_handler, set_key_handler, set_timer,
                           event_loop)


# adjust these "constants" as you see fit
DEFAULT_ROWS = 40
DEFAULT_COLUMNS = 60
SCALE = 15
ACTIVE_COLOR = 'green'
INACTIVE_COLOR = 'yellow'
# ...


# keep use of global variables to minimum
_board = None
# ...


def gol(columns=DEFAULT_COLUMNS, rows=DEFAULT_ROWS):
    global _board
    _board = build_grid(columns, rows, False)
    open_window(columns=columns, rows=rows, scale=SCALE,
                show_cell_borders=True,
                title="Conway's Game of Life")
    paint_grid(_board)
    set_key_handler(main_key_handler)
    # set_click_handler(click_handler)
    event_loop()


# supplied for you
def get_cell(grid, p):
    """
    Returns contents of grid based at point p. Assumes p consists of a valid
    column and row in grid.

    >>> g = [['A', 'B', 'C'], ['D', 'E', 'F']]
    >>> get_cell(g, (2, 1))
    'F'
    """
    (x, y) = p
    v = grid[y][x]
    return v


# 1.
def set_cell(grid, p, value=True):
    """
    Stores value at coordinate point p in grid. Assumes p consists of a valid
    column and row in grid.

    >>> g = [['A', 'B', 'C'], ['D', 'E', 'F']]
    >>> set_cell(g, (2, 0), 'X')
    >>> g
    [['A', 'B', 'X'], ['D', 'E', 'F']]
    """
    ...


# provided for you
def set_cells(grid, points, value=True):
    for p in points:
        set_cell(grid, p, value)


# provided for you
def update_cell(grid, p):
    if get_cell(grid, p):
        paint(p, ACTIVE_COLOR)
    else:
        paint(p, INACTIVE_COLOR)


# provided for you
def update_cells(grid, points):
    for p in points:
        update_cell(grid, p)


# 2. Modifies cell at coordinate p in grid to be inverse of what is was (False
#     becomes True and vice versa). In addition, the corresponding coordinate
#     box in the graphics window is painted in the graphics window.
#     call: get_cell, set_cell, update_cell
def toggle_cell(grid, p):
    ...


# provided for you
def toggle_board():
    (columns, rows) = grid_size(_board)
    for y in range(rows):
        for x in range(columns):
            p = (x, y)
            toggle_cell(_board, p)


# 3. ... define the click handler ...


# --------------------------------------------
# 4.
def count_neighbors(grid, p):
    count = 0
    ...
    return count


# 5. call count_neighbors, get_cell
def compute_changes(grid):
    births = []
    deaths = []
    ...
    return (births, deaths)


# 6. call compute_changes, set_cells, update_cells
def evolve_once():
    ...


# 7. call evolve_once, set_timer
def evolve_loop():
    ...


# 8.
def start_evolution():
    ...
    evolve_loop()
    ...


# ...


# 9. assumes valid filename
#    use .split, build_grid, set_cell
def read_gridfile(filename):
    f = open(filename)
    text = f.read()
    f.close()
    ...
    grid = build_grid(..., ..., False)
    ...
    return grid


# provided for you
def graft_grid(src_grid, dest_grid):
    (sc, sr) = grid_size(src_grid)
    (dc, dr) = grid_size(dest_grid)
    for y in range(min(sr, dr)):
        for x in range(min(sc, dc)):
            p = (x, y)
            set_cell(dest_grid, p, get_cell(src_grid, p))


def load_file():
    filename = input('Enter name of GOL grid file to load: ').strip()
    g = read_gridfile(filename)
    graft_grid(g, _board)
    paint_grid(_board)


# 10.
def census(grid):
    population = 0
    ...
    return population


def print_stats(grid):
    ...
    print('total population: ...')
    ...


# 11.
def randomize_grid(grid, density):
    ...


# ...


# keep this function as simple as possible
# generally have each if/elif case just call one function
def main_key_handler(k):
    if k == 'q':
        close_window()
    elif k == 'h':
        print_help()
    elif k == 'c':
        fill_grid(_board, False)
        paint_grid(_board)
    elif k == 'f':
        fill_grid(_board, True)
        paint_grid(_board)
    elif k == 't':
        toggle_board()
    elif k == 'o' or k == 'space':
        evolve_once()
    elif k == 'e':
        start_evolution()
    elif k == 'l':
        load_file()
    elif k == 's':
        print_stats(_board)
    # ...
    else:
        print(f'key "{k}" has no effect here')


# ... add to the menu as necessary ...
def print_help():
    print("""
Game of Life: Main Menu
=======================
c)lear board
f)ill board
t)oggle board

o)ne step of evolution (or space bar)
e)volve continually

l)oad grid from file

h)elp: display this menu
q)uit
""")


# -----------------------------------------------------------------------
# helpful functions supplied for you


def build_grid(columns, rows, initial_value):
    grid = []
    for _ in range(rows):
        row = []
        for _ in range(columns):
            row.append(initial_value)
        grid.append(row)
    return grid


def grid_size(grid):
    """
    Returns a pair of integers indicating the width (number of columns)
    and height (number of rows) in grid. Assumes grid is a list of lists
    where each interior list of the same length.
    """
    rows = len(grid)
    if rows > 0:
        columns = len(grid[0])
    else:
        columns = 0
    return (columns, rows)


def fill_grid(grid, value):
    (columns, rows) = grid_size(grid)
    for y in range(rows):
        for x in range(columns):
            p = (x, y)
            set_cell(grid, p, value)


# this departs a small bit from the way we did this in class: we do not assume
# that the grid that is being drawn is necessarily the same size as the
# graphics window and we paint all the squares where on/off (True/False)
# indicates which color box should appear
def paint_grid(grid):
    (gc, gr) = grid_size(grid)
    (wc, wr) = get_dimensions()
    clear()
    for y in range(min(gr, wr)):
        for x in range(min(gc, wc)):
            p = (x, y)
            if get_cell(grid, p):
                paint(p, ACTIVE_COLOR)
            else:
                paint(p, INACTIVE_COLOR)
    update()   # <--- necessary since might get called outside of event loop


# -----------------------------------------------------------------------
# game will automatically be started if this module loaded as main module

if __name__ == '__main__':
    gol()
